package BAR_GRAPH;

# +-------------------------------------------------------------------+
# |                   H T M L - G R A P H S   (v4.10)                 |
# |                                                                   |
# | Copyright Gerd Tentler                www.gerd-tentler.de/tools   |
# | Created: Sep. 17, 2002                Last modified: Jan. 7, 2012 |
# +-------------------------------------------------------------------+
# | This program may be used and hosted free of charge by anyone for  |
# | personal purpose as long as this copyright notice remains intact. |
# |                                                                   |
# | Obtain permission before selling the code for this program or     |
# | hosting this software on a commercial website or redistributing   |
# | this software over the Internet or in any other medium. In all    |
# | cases copyright must remain intact.                               |
# +-------------------------------------------------------------------+
#
#======================================================================================================
# Example:
#
#   require 'BAR_GRAPH.pm';
#   $graph = BAR_GRAPH->new("hBar");
#   $graph->{values} = [234, 125, 289, 147, 190];
#   print $graph->create();
#
# Returns HTML code
#======================================================================================================
#

#------------------------------------------------------------------------------------------------------
# Class Methods
#------------------------------------------------------------------------------------------------------

sub new {
	my $type = $_[1];
	my $self = {};
	bless $self;
	$self->initialize();
	if($type) { $self->{type} = $type; }
	return $self;
}

sub initialize {
	my $this = shift;
	#----------------------------------------------------------------------------------------------------
	# Configuration
	#----------------------------------------------------------------------------------------------------
	$this->{type} = 'hBar';                        # graph type: "hBar", "vBar", "pBar", or "fader"
	$this->{values} = undef;                       # graph data: array or string with comma-separated values

	$this->{graphBGColor} = '';                    # graph background color: string
	$this->{graphBorder} = '';                     # graph border: string (CSS specification; doesn't work with NN4)
	$this->{graphPadding} = 0;                     # graph padding: integer (pixels)

	$this->{titles} = undef;                       # titles: array or string with comma-separated values
	$this->{titleColor} = 'black';                 # title font color: string
	$this->{titleBGColor} = '#C0E0FF';             # title background color: string
	$this->{titleBorder} = '2px groove white';     # title border: string (CSS specification)
	$this->{titleFont} = 'Arial, Helvetica';       # title font family: string (CSS specification)
	$this->{titleSize} = 12;                       # title font size: integer (pixels)
	$this->{titleAlign} = 'center';                # title text align: "left", "center", or "right"
	$this->{titlePadding} = 2;                     # title padding: integer (pixels)

	$this->{labels} = undef;                       # label names: array or string with comma-separated values
	$this->{labelColor} = 'black';                 # label font color: string
	$this->{labelBGColor} = '#C0E0FF';             # label background color: string
	$this->{labelBorder} = '2px groove white';     # label border: string (CSS specification; doesn't work with NN4)
	$this->{labelFont} = 'Arial, Helvetica';       # label font family: string (CSS specification)
	$this->{labelSize} = 12;                       # label font size: integer (pixels)
	$this->{labelAlign} = 'center';                # label text align: "left", "center", or "right"
	$this->{labelSpace} = 0;                       # additional space between labels: integer (pixels)

	$this->{barWidth} = 20;                        # bar width: integer (pixels)
	$this->{barLength} = 1.0;                      # bar length ratio: float (from 0.1 to 2.9)
	$this->{barColors} = undef;                    # bar colors OR bar images: array or string with comma-separated values
	$this->{barBGColor} = undef;                   # bar background color: string
	$this->{barBorder} = '2px outset white';       # bar border: string (CSS specification; doesn't work with NN4)
	$this->{barLevelColors} = undef;               # bar level colors: ascending array (bLevel, bColor[,...]); draw bars >= bLevel with bColor

	$this->{showValues} = 0;                       # show values: 0 = % only, 1 = abs. and %, 2 = abs. only, 3 = none
	$this->{baseValue} = 0;                        # base value: integer or float (only hBar and vBar)

	$this->{absValuesColor} = 'black';             # abs. values font color: string
	$this->{absValuesBGColor} = '#C0E0FF';         # abs. values background color: string
	$this->{absValuesBorder} = '2px groove white'; # abs. values border: string (CSS specification; doesn't work with NN4)
	$this->{absValuesFont} = 'Arial, Helvetica';   # abs. values font family: string (CSS specification)
	$this->{absValuesSize} = 12;                   # abs. values font size: integer (pixels)
	$this->{absValuesPrefix} = '';                 # abs. values prefix (e.g. "$")
	$this->{absValuesSuffix} = '';                 # abs. values suffix (e.g. " kg")

	$this->{percValuesColor} = 'black';            # perc. values font color: string
	$this->{percValuesFont} = 'Arial, Helvetica';  # perc. values font family: string (CSS specification)
	$this->{percValuesSize} = 12;                  # perc. values font size: integer (pixels)
	$this->{percValuesDecimals} = 0;               # perc. values number of decimals: integer

	$this->{charts} = 1;                           # number of charts: integer

	# hBar/vBar only:
	$this->{legend} = undef;                       # legend items: array or string with comma-separated values
	$this->{legendColor} = 'black';                # legend font color: string
	$this->{legendBGColor} = '#F0F0F0';            # legend background color: string
	$this->{legendBorder} = '2px groove white';    # legend border: string (CSS specification; doesn't work with NN4)
	$this->{legendFont} = 'Arial, Helvetica';      # legend font family: string (CSS specification)
	$this->{legendSize} = 12;                      # legend font size: integer (pixels)
	$this->{legendAlign} = 'top';                  # legend vertical align: "top", "center", "bottom"
	$this->{legendAbsValues} = 0;                  # view abs. values in legend: 1 = yes, 0 = no

	# debug mode: 0 = off, 1 = on; just shows some extra information
	$this->{debug} = 0;

	# default bar colors; only used if my barColors isn't set
	$this->{colors} = ['#0000FF', '#FF0000', '#00E000', '#A0A0FF', '#FFA0A0', '#00A000'];

	# error messages
	$this->{err_type} = 'ERROR: Type must be "hBar", "vBar", "pBar", or "fader"';

	# CSS names (don't change)
	$this->{cssGRAPH} = '';
	$this->{cssBAR} = '';
	$this->{cssBARBG} = '';
	$this->{cssTITLE} = '';
	$this->{cssLABEL} = '';
	$this->{cssLABELBG} = '';
	$this->{cssLEGEND} = '';
	$this->{cssLEGENDBG} = '';
	$this->{cssABSVALUES} = '';
	$this->{cssPERCVALUES} = '';
	#----------------------------------------------------------------------------------------------------
}

sub set_styles {
	my $this = shift;
	if($this->{graphBGColor}) { $this->{cssGRAPH} .= 'background-color:' . $this->{graphBGColor} . ';'; }
	if($this->{graphBorder}) { $this->{cssGRAPH} .= 'border:' . $this->{graphBorder} . ';'; }
	if($this->{barBorder}) { $this->{cssBAR} .= 'border:' . $this->{barBorder} . ';'; }
	if($this->{barBGColor}) { $this->{cssBARBG} .= 'background-color:' . $this->{barBGColor} . ';'; }
	if($this->{titleColor}) { $this->{cssTITLE} .= 'color:' . $this->{titleColor} . ';'; }
	if($this->{titleBGColor}) { $this->{cssTITLE} .= 'background-color:' . $this->{titleBGColor} . ';'; }
	if($this->{titleBorder}) { $this->{cssTITLE} .= 'border:' . $this->{titleBorder} . ';'; }
	if($this->{titleFont}) { $this->{cssTITLE} .= 'font-family:' . $this->{titleFont} . ';'; }
	if($this->{titleAlign}) { $this->{cssTITLE} .= 'text-align:' . $this->{titleAlign} . ';'; }
	if($this->{titleSize}) { $this->{cssTITLE} .= 'font-size:' . $this->{titleSize} . 'px;'; }
	if($this->{titleBGColor}) { $this->{cssTITLE} .= 'background-color:' . $this->{titleBGColor} . ';'; }
	if($this->{titlePadding}) { $this->{cssTITLE} .= 'padding:' . $this->{titlePadding} . 'px;'; }
	if($this->{labelColor}) { $this->{cssLABEL} .= 'color:' . $this->{labelColor} . ';'; }
	if($this->{labelBGColor}) { $this->{cssLABEL} .= 'background-color:' . $this->{labelBGColor} . ';'; }
	if($this->{labelBorder}) { $this->{cssLABEL} .= 'border:' . $this->{labelBorder} . ';'; }
	if($this->{labelFont}) { $this->{cssLABEL} .= 'font-family:' . $this->{labelFont} . ';'; }
	if($this->{labelSize}) { $this->{cssLABEL} .= 'font-size:' . $this->{labelSize} . 'px;'; }
	if($this->{labelAlign}) { $this->{cssLABEL} .= 'text-align:' . $this->{labelAlign} . ';'; }
	if($this->{labelBGColor}) { $this->{cssLABELBG} .= 'background-color:' . $this->{labelBGColor} . ';'; }
	if($this->{legendColor}) { $this->{cssLEGEND} .= 'color:' . $this->{legendColor} . ';'; }
	if($this->{legendFont}) { $this->{cssLEGEND} .= 'font-family:' . $this->{legendFont} . ';'; }
	if($this->{legendSize}) { $this->{cssLEGEND} .= 'font-size:' . $this->{legendSize} . 'px;'; }
	if($this->{legendBGColor}) { $this->{cssLEGENDBG} .= 'background-color:' . $this->{legendBGColor} . ';'; }
	if($this->{legendBorder}) { $this->{cssLEGENDBG} .= 'border:' . $this->{legendBorder} . ';'; }
	if($this->{absValuesColor}) { $this->{cssABSVALUES} .= 'color:' . $this->{absValuesColor} . ';'; }
	if($this->{absValuesBGColor}) { $this->{cssABSVALUES} .= 'background-color:' . $this->{absValuesBGColor} . ';'; }
	if($this->{absValuesBorder}) { $this->{cssABSVALUES} .= 'border:' . $this->{absValuesBorder} . ';'; }
	if($this->{absValuesFont}) { $this->{cssABSVALUES} .= 'font-family:' . $this->{absValuesFont} . ';'; }
	if($this->{absValuesSize}) { $this->{cssABSVALUES} .= 'font-size:' . $this->{absValuesSize} . 'px;'; }
	if($this->{percValuesColor}) { $this->{cssPERCVALUES} .= 'color:' . $this->{percValuesColor} . ';'; }
	if($this->{percValuesFont}) { $this->{cssPERCVALUES} .= 'font-family:' . $this->{percValuesFont} . ';'; }
	if($this->{percValuesSize}) { $this->{cssPERCVALUES} .= 'font-size:' . $this->{percValuesSize} . 'px;'; }
}

sub level_color {
	my ($this, $value, $color) = @_;
	my ($i, @barLevelColors);
	if($this->{barLevelColors}) {
		@barLevelColors = @{$this->{barLevelColors}};
		for($i = 0; $i < int @barLevelColors; $i += 2) {
			if($i+1 < int @barLevelColors) {
				if(($barLevelColors[$i] > 0 && $value >= $barLevelColors[$i]) ||
					($barLevelColors[$i] < 0 && $value <= $barLevelColors[$i])) {
					$color = $barLevelColors[$i+1];
				}
			}
		}
	}
	return $color;
}

sub build_bar {
	my ($this, $value, $width, $height, $color) = @_;
	my $title = $this->{absValuesPrefix} . $value . $this->{absValuesSuffix};
	my $bg = ($color =~ /\.(jpg|jpeg|jpe|gif|png)$/i) ? 'background' : 'bgcolor';
	my $bar = '<table border=0 cellspacing=0 cellpadding=0><tr>';
	$bar .= '<td style="' . $this->{cssBAR} . '" ' . $bg . '="' . $color . '"';
	$bar .= ($value ne '') ? ' title="' . $title . '">' : '>';
	$bar .= '<div style="width:' . $width . 'px; height:' . $height . 'px;';
	$bar .= ' line-height:1px; font-size:1px;"></div>';
	$bar .= '</td></tr></table>';
	return $bar;
}

sub build_fader {
	my ($this, $value, $width, $height, $x, $color) = @_;
	my $fader = '<table border=0 cellspacing=0 cellpadding=0><tr>';
	$x -= _round($width / 2);
	if($x > 0) { $fader .= '<td width=' . $x . '></td>'; }
	$fader .= '<td>' . $this->build_bar($value, $width, $height, $color) . '</td>';
	$fader .= '</tr></table>';
	return $fader;
}

sub build_value {
	my ($this, $val, $max_dec, $sum, $align) = @_;
	$val = _number_format($val, $max_dec);
	if($sum) { $sum = _number_format($sum, $max_dec); }
	my $value = '<td style="' . $this->{cssABSVALUES} . '"';
	if($align) { $value .= ' align=' . $align; }
	$value .= ' nowrap>';
	$value .= '&nbsp;' . $this->{absValuesPrefix} . $val . $this->{absValuesSuffix};
	if($sum) { $value .= ' / ' . $this->{absValuesPrefix} . $sum . $this->{absValuesSuffix}; }
	$value .= '&nbsp;</td>';
	return $value;
}

sub build_legend {
	my $this = $_[0];
	my @barColors = @{$_[1]};
	my @barValues = @{$_[2]};
	my $legend = '<table border=0 cellspacing=0 cellpadding=0><tr>';
	$legend .= '<td style="' . $this->{cssLEGENDBG} . '">';
	$legend .= '<table border=0 cellspacing=4 cellpadding=0>';
	my @l = (_is_array($this->{legend})) ? @{$this->{legend}} : split(',', $this->{legend});
	my ($i, $chart, $label, $sum) = 0;

	for(; $i < int @barColors; $i++) {
		$legend .= '<tr>';
		$legend .= '<td>' . $this->build_bar('', $this->{barWidth}, $this->{barWidth}, $barColors[$i]) . '</td>';
		$legend .= '<td style="' . $this->{cssLEGEND} . '" nowrap>' . _trim($l[$i]) . '</td>';

		if($this->{legendAbsValues}) {
			for($chart = $sum = 0; $chart < int @barValues; $chart++) {
				for($label = 0; $label < int @{$barValues[$chart]}; $label++) {
					$sum += $barValues[$chart][$label][$i];
				}
			}
			$legend .= '<td style="' . $this->{cssLEGEND} . '" nowrap>(' . $this->{absValuesPrefix} . $sum . $this->{absValuesSuffix} . ')</td>';
		}
		$legend .= '</tr>';
	}
	$legend .= '</table></td></tr></table>';
	return $legend;
}

sub build_hTitle {
	my ($this, $titleLabel, $titleValue, $titleBar) = @_;
	my $title = '<tr>';
	$title .= '<td style="' . $this->{cssTITLE} . '">' . $titleLabel . '</td>';
	if($titleValue ne '') { $title .= '<td style="' . $this->{cssTITLE} . '">' . $titleValue . '</td>'; }
	$title .= '<td style="' . $this->{cssTITLE} . '">' . $titleBar . '</td>';
	$title .= '</tr>';
	return $title;
}

sub create_hBar {
	my ($this, $value, $percent, $mPerc, $mPerc_neg, $max_neg, $mul, $valSpace, $bColor, $border, $spacer, $spacer_neg) = @_;
	my $bar = '<table border=0 cellspacing=0 cellpadding=0 height=100%><tr>';

	if($percent < 0) {
		$percent *= -1;
		$bar .= '<td style="' . $this->{cssLABELBG} . '" height=' . $this->{barWidth} . ' width=' . _round(($mPerc_neg - $percent) * $mul + $valSpace) . ' align=right nowrap>';
		if($this->{showValues} < 2) { $bar .= '<span style="' . $this->{cssPERCVALUES} . '">' . _number_format($percent, $this->{percValuesDecimals}) . '%</span>'; }
		$bar .= '&nbsp;</td><td style="' . $this->{cssLABELBG} . '">';
		$bar .= $this->build_bar($value, _round($percent * $mul), $this->{barWidth}, $bColor);
		$bar .= '</td><td width=' . $spacer . '></td>';
	}
	else {
		if($max_neg) {
			$bar .= '<td style="' . $this->{cssLABELBG} . '" width=' . $spacer_neg . '>';
			$bar .= '<table border=0 cellspacing=0 cellpadding=0><tr><td></td></tr></table></td>';
		}
		if($percent) {
			$bar .= '<td>';
			$bar .= $this->build_bar($value, _round($percent * $mul), $this->{barWidth}, $bColor);
			$bar .= '</td>';
		}
		else { $bar .= '<td width=1 height=' . ($this->{barWidth} + ($border * 2)) . '></td>'; }
		$bar .= '<td style="' . $this->{cssPERCVALUES} . '" width=' . _round(($mPerc - $percent) * $mul + $valSpace) . ' align=left nowrap>';
		if($this->{showValues} < 2) { $bar .= '&nbsp;' . _number_format($percent, $this->{percValuesDecimals}) . '%'; }
		$bar .= '&nbsp;</td>';
	}
	$bar .= '</tr></table>';

	return $bar;
}

sub create_vBar {
	my ($this, $value, $percent, $mPerc, $mPerc_neg, $max_neg, $mul, $valSpace, $bColor, $border, $spacer, $spacer_neg) = @_;
	my $bar = '<table border=0 cellspacing=0 cellpadding=0 width=100%><tr align=center>';

	if($percent < 0) {
		$percent *= -1;
		$bar .= '<td height=' . $spacer . '></td></tr><tr align=center valign=top><td style="' . $this->{cssLABELBG} . '">';
		$bar .= $this->build_bar($value, $this->{barWidth}, _round($percent * $mul), $bColor);
		$bar .= '</td></tr><tr align=center valign=top>';
		$bar .= '<td style="' . $this->{cssLABELBG} . '" height=' . _round(($mPerc_neg - $percent) * $mul + $valSpace) . ' nowrap>';
		$bar .= ($this->{showValues} < 2) ? '<span style="' . $this->{cssPERCVALUES} . '">' . _number_format($percent, $this->{percValuesDecimals}) . '%</span>' : '&nbsp;';
		$bar .= '</td>';
	}
	else {
		$bar .= '<td style="' . $this->{cssPERCVALUES} . '" valign=bottom height=' . _round(($mPerc - $percent) * $mul + $valSpace) . ' nowrap>';
		if($this->{showValues} < 2) { $bar .= _number_format($percent, $this->{percValuesDecimals}) . '%'; }
		$bar .= '</td>';
		if($percent) {
			$bar .= '</tr><tr align=center valign=bottom><td>';
			$bar .= $this->build_bar($value, $this->{barWidth}, _round($percent * $mul), $bColor);
			$bar .= '</td>';
		}
		else { $bar .= '</tr><tr><td width=' . ($this->{barWidth} + ($border * 2)) . ' height=1></td>'; }
		if($max_neg) {
			$bar .= '</tr><tr><td style="' . $this->{cssLABELBG} . '" height=' . $spacer_neg . '>';
			$bar .= '<table border=0 cellspacing=0 cellpadding=0><tr><td></td></tr></table></td>';
		}
	}
	$bar .= '</tr></table>';

	return $bar;
}

sub create {
	my $this = shift;
	my ($r, $graph, $label, $rowspan, $colspan, $bColor, $titleLabel, $titleValue, $titleBar) = '';
	my (@d, @t, @r, @drc, @drv, @bc, @val) = ();
	my ($bars, $divide, $border, $mPerc, $mul, $valSpace, $spacer_neg, $spacer, $mPerc_neg, $maxSize, $percent) = 0;
	my ($i, $j, $sum, $max, $max_neg, $max_dec, $ccnt, $lcnt, $chart, $dec, $v) = 0;
	my @colors = @{$this->{colors}};
	my @barLevelColors = @{$this->{barLevelColors}};

	$this->{type} = lc($this->{type});
	@d = (_is_array($this->{values})) ? @{$this->{values}} : split(',', $this->{values});
	if(_is_array($this->{titles})) { @t = @{$this->{titles}}; }
	else { @t = (length($this->{titles}) > 1) ? split(',', $this->{titles}) : (); }
	if(_is_array($this->{labels})) { @r = @{$this->{labels}}; }
	else { @r = (length($this->{labels}) > 1) ? split(',', $this->{labels}) : (); }
	if($this->{barColors}) { @drc = (_is_array($this->{barColors})) ? @{$this->{barColors}} : split(',', $this->{barColors}); }
	else { @drc = (); }
	if($this->{barLength} < 0.1) { $this->{barLength} = 0.1; }
	elsif($this->{barLength} > 2.9) { $this->{barLength} = 2.9; }
	$bars = (int @d > int @r) ? int @d : int @r;

	if($this->{type} eq 'pbar' || $this->{type} eq 'fader') {
		if(!$this->{barBGColor}) { $this->{barBGColor} = $this->{labelBGColor}; }
		if($this->{labelBGColor} eq $this->{barBGColor} && int @t == 0) {
			$this->{labelBGColor} = '';
			$this->{labelBorder} = '';
		}
	}

	$this->set_styles();

	$graph .= '<table border=0 cellspacing=0 cellpadding=' . $this->{graphPadding} . '><tr>';
	$graph .= '<td style="' . $this->{cssGRAPH} . '">';

	if($this->{legend} && $this->{type} ne 'pbar' && $this->{type} ne 'fader') {
		$graph .= '<table border=0 cellspacing=0 cellpadding=0><tr><td>';
	}
	if($this->{charts} > 1) {
		$divide = _ceil($bars / $this->{charts});
		$graph .= '<table border=0 cellspacing=0 cellpadding=6><tr valign=top><td>';
	}
	else { $divide = 0; }

	for(($i, $sum, $max, $max_neg, $max_dec, $ccnt, $lcnt, $chart) = 0; $i < $bars; $i++) {
		if($divide && $i && !($i % $divide)) {
			$lcnt = 0;
			$chart++;
		}
		@drv = split(';', $d[$i]);

		for(($j, $dec) = 0; $j < int @drv; $j++) {
			$val[$chart][$lcnt][$j] = _trim($drv[$j]);
			$v = $val[$chart][$lcnt][$j] ? $val[$chart][$lcnt][$j] - $this->{baseValue} : 0;

			if($this->{type} eq 'hbar' || $this->{type} eq 'vbar' || $j == 0) {
				if($v > $max) { $max = $v; }
				elsif($v < $max_neg) { $max_neg = $v; }
			}

			if($v < 0) { $v *= -1; }
			$sum += $v;

			if($v =~ /\./) {
				$v = _round($v * 1000000000000) / 1000000000000;
				$dec = length(substr($v, rindex($v, '.') + 1));
				if($dec > $max_dec) { $max_dec = $dec; }
			}

			if(!$bc[$j]) {
				if($ccnt >= int @colors) { $ccnt = 0; }
				$bc[$j] = (!$drc[$j] || length($drc[$j]) < 3) ? @colors[$ccnt++] : _trim($drc[$j]);
			}
		}
		$lcnt++;
	}

	if(@barLevelColors) {
		for($i = 0; $i < int @barLevelColors; $i += 2) {
			if(@barLevelColors[$i] == 'MAX') { $this->{barLevelColors}[$i] = $max; }
		}
	}

	$border = int($this->{barBorder});
	$mPerc = $sum ? _round($max * 100 / $sum) : 0;
	if($this->{type} eq 'pbar' || $this->{type} eq 'fader') { $mul = 2; }
	else { $mul = $mPerc ? 100 / $mPerc : 1; }
	$mul *= $this->{barLength};

	if($this->{showValues} < 2) {
		if($this->{type} eq 'hbar') {
			$valSpace = ($this->{percValuesDecimals} * ($this->{percValuesSize} / 1.6)) + ($this->{percValuesSize} * 3.2);
		}
		else { $valSpace = $this->{percValuesSize} * 1.2; }
	}
	else { $valSpace = $this->{percValuesSize}; }
	$maxSize = _round($mPerc * $mul + $valSpace + $border * 2);
	$spacer = $maxSize;

	if($max_neg) {
		$mPerc_neg = $sum ? _round(-$max_neg * 100 / $sum) : 0;
		if($mPerc_neg > $mPerc && $this->{type} ne 'pbar' && $this->{type} ne 'fader') {
			$mul = 100 / $mPerc_neg * $this->{barLength};
		}
		$spacer_neg = _round($mPerc_neg * $mul + $valSpace + $border * 2);
		$maxSize += $spacer_neg;
	}

	if(int @t > 0) {
		$titleLabel = ($t[0] eq '') ? '&nbsp;' : $t[0];

		if($this->{showValues} == 1 || $this->{showValues} == 2) {
			$titleValue = ($t[1] eq '') ? '&nbsp;' : $t[1];
			$titleBar = ($t[2] eq '') ? '&nbsp;' : $t[2];
		}
		else { $titleBar = ($t[1] eq '') ? '&nbsp;' : $t[1]; }
	}

	for(($chart, $lcnt) = 0; $chart < int @val; $chart++) {
		$graph .= '<table border=0 cellspacing=2 cellpadding=0>';

		if($this->{type} eq 'hbar') {
			if(int @t > 0) { $graph .= $this->build_hTitle($titleLabel, $titleValue, $titleBar); }

			for($i = 0; $i < int @{$val[$chart]}; $i++, $lcnt++) {
				$label = ($lcnt < int @r) ? _trim($r[$lcnt]) : $lcnt+1;
				$rowspan = int @{$val[$chart][$i]};
				$graph .= '<tr><td style="' . $this->{cssLABEL} . '"' . (($rowspan > 1) ? ' rowspan=' . $rowspan : '') . '>';
				$graph .= '&nbsp;' . $label . '&nbsp;</td>';

				for($j = 0; $j < int @{$val[$chart][$i]}; $j++) {
					$value = $val[$chart][$i][$j] ? $val[$chart][$i][$j] - $this->{baseValue} : 0;
					$percent = $sum ? $value * 100 / $sum : 0;
					$value = _number_format($val[$chart][$i][$j], $max_dec);
					$bColor = $this->level_color($val[$chart][$i][$j], $bc[$j]);

					if($this->{showValues} == 1 || $this->{showValues} == 2) {
						$graph .= $this->build_value($val[$chart][$i][$j], $max_dec, 0, 'right');
					}
					$graph .= '<td' . ($this->{cssBARBG} ? ' style="' . $this->{cssBARBG} . '"' : '') . ' height=100% width=' . $maxSize . '>';
					$graph .= $this->create_hBar($value, $percent, $mPerc, $mPerc_neg, $max_neg, $mul, $valSpace, $bColor, $border, $spacer, $spacer_neg);
					$graph .= '</td></tr>';
					if($j < int @{$val[$chart][$i]} - 1) { $graph .= '<tr>'; }
				}
				if($this->{labelSpace} && $i < int @{$val[$chart]}-1) { $graph .= '<tr><td colspan=3 height=' . $this->{labelSpace} . '></td></tr>'; }
			}
		}
		elsif($this->{type} eq 'vbar') {
			$graph .= '<tr align=center valign=bottom>';

			if($titleBar ne '') {
				$titleBar =~ s/-/-<br>/g;
				$graph .= '<td style="' . $this->{cssTITLE} . '" valign=middle>' . $titleBar . '</td>';
			}
			for($i = 0; $i < int @{$val[$chart]}; $i++) {
				for($j = 0; $j < int @{$val[$chart][$i]}; $j++) {
					$value = $val[$chart][$i][$j] ? $val[$chart][$i][$j] - $this->{baseValue} : 0;
					$percent = $sum ? $value * 100 / $sum : 0;
					$value = _number_format($val[$chart][$i][$j], $max_dec);
					$bColor = $this->level_color($val[$chart][$i][$j], $bc[$j]);

					$graph .= '<td' . ($this->{cssBARBG} ? ' style="' . $this->{cssBARBG} . '"' : '') . '>';
					$graph .= $this->create_vBar($value, $percent, $mPerc, $mPerc_neg, $max_neg, $mul, $valSpace, $bColor, $border, $spacer, $spacer_neg);
					$graph .= '</td>';
				}
				if($this->{labelSpace}) { $graph .= '<td width=' . $this->{labelSpace} . '></td>'; }
			}
			if($this->{showValues} == 1 || $this->{showValues} == 2) {
				$graph .= '</tr><tr align=center>';
				if($titleValue ne '') { $graph .= '<td style="' . $this->{cssTITLE} . '">' . $titleValue . '</td>'; }

				for($i = 0; $i < int @{$val[$chart]}; $i++) {
					for($j = 0; $j < int @{$val[$chart][$i]}; $j++) {
						$graph .= $this->build_value($val[$chart][$i][$j], $max_dec);
					}
					if($this->{labelSpace}) { $graph .= '<td width=' . $this->{labelSpace} . '></td>'; }
				}
			}
			$graph .= '</tr><tr>';
			if($titleLabel ne '') { $graph .= '<td style="' . $this->{cssTITLE} . '">' . $titleLabel . '</td>'; }

			for($i = 0; $i < int @{$val[$chart]}; $i++, $lcnt++) {
				$label = ($lcnt < int @r) ? _trim($r[$lcnt]) : $lcnt+1;
				$colspan = int @{$val[$chart][$i]};
				$graph .= '<td style="' . $this->{cssLABEL} . '"' . (($colspan > 1) ? ' colspan=' . $colspan : '') . '>';
				$graph .= '&nbsp;' . $label . '&nbsp;</td>';
				if($this->{labelSpace}) { $graph .= '<td width=' . $this->{labelSpace} . '></td>'; }
			}
			$graph .= '</tr>';
		}
		elsif($this->{type} eq 'pbar' || $this->{type} eq 'fader') {
			if(int @t > 0) { $graph .= $this->build_hTitle($titleLabel, $titleValue, $titleBar); }

			for($i = 0; $i < int @{$val[$chart]}; $i++, $lcnt++) {
				$label = ($lcnt < int @r) ? _trim($r[$lcnt]) : '';
				$graph .= '<tr>';

				if($label) {
					$graph .= '<td style="' . $this->{cssLABEL} . '">';
					$graph .= '&nbsp;' . $label . '&nbsp;</td>';
				}
				$sum = $val[$chart][$i][1];
				$percent = $sum ? $val[$chart][$i][0] * 100 / $sum : 0;
				$value = _number_format($val[$chart][$i][0], $max_dec);

				if($this->{showValues} == 1 || $this->{showValues} == 2) {
					$graph .= $this->build_value($val[$chart][$i][0], $max_dec, $sum, 'right');
				}
				$graph .= '<td' . ($this->{cssBARBG} ? ' style="' . $this->{cssBARBG} . '"' : '') . '>';

				$this->{barColors} = $drc[$i] ? _trim($drc[$i]) : @colors[0];
				$bColor = $this->level_color($val[$chart][$i][0], $this->{barColors});
				$graph .= '<table border=0 cellspacing=0 cellpadding=0><tr><td>';
				if($this->{type} eq 'fader') { $graph .= $this->build_fader($value, _round($this->{barWidth} / 2), $this->{barWidth}, _round($percent * $mul), $bColor); }
				else { $graph .= $this->build_bar($value, _round($percent * $mul), $this->{barWidth}, $bColor); }
				$graph .= '</td><td width=' . _round((100 - $percent) * $mul) . '></td>';
				$graph .= '</tr></table></td>';
				if($this->{showValues} < 2) {
					$graph .= '<td style="' . $this->{cssPERCVALUES} . '" nowrap>&nbsp;' . _number_format($percent, $this->{percValuesDecimals}) . '%</td>';
				}
				$graph .= '</tr>';
				if($this->{labelSpace} && $i < int @{$val[$chart]}-1) { $graph .= '<td colspan=3 height=' . $this->{labelSpace} . '></td>'; }
			}
		}
		else { $graph .= '<tr><td>' . $this->{err_type} . '</td></tr>'; }

		$graph .= '</table>';

		if($chart < $this->{charts} - 1 && int $val[$chart+1]) {
			$graph .= '</td>';
			if($this->{type} eq 'vbar') { $graph .= '</tr><tr valign=top>'; }
			$graph .= '<td>';
		}
	}

	if($this->{charts} > 1) { $graph .= '</td></tr></table>'; }

	if($this->{legend} && $this->{type} ne 'pbar' && $this->{type} ne 'fader') {
		$graph .= '</td><td width=10>&nbsp;</td><td' . ($this->{legendAlign} ? ' valign=' . $this->{legendAlign} : '') . '>';
		$graph .= $this->build_legend(\@bc, \@val);
		$graph .= '</td></tr></table>';
	}

	if($this->{debug}) {
		$graph .= "<br>sum=$sum max=$max max_neg=$max_neg max_dec=$max_dec ";
		$graph .= "mPerc=$mPerc mPerc_neg=$mPerc_neg mul=$mul valSpace=$valSpace ";
		$graph .= "spacer=$spacer spacer_neg=$spacer_neg";
	}

	$graph .= '</td></tr></table>';

	return $graph;
}

#------------------------------------------------------------------------------------------------------
# Global Functions
#------------------------------------------------------------------------------------------------------

sub _trim {
	my $s = shift;
	$s =~ s/^\s+|\s+$//g;
	return $s;
}

sub _is_array {
	my $a = shift;
	return (ref($a) eq 'ARRAY') ? 1 : 0;
}

sub _round {
	my $fl = shift;
	return ($fl >= 0) ? int($fl + 0.5) : int($fl - 0.5);
}

sub _ceil {
	my $fl = shift;
	return int($fl + 0.9);
}

sub _number_format {
	my ($val, $dec) = @_;
	return $dec ? sprintf('%.' . $dec . 'f', $val) : _round($val);
}

#------------------------------------------------------------------------------------------------------

return 1;
